@@ -6,19 +6,16 @@ from decimal import Decimal
 
 from sql import Column
 
+from trytond.cache import Cache
 from trytond.i18n import gettext
 from trytond.model import (
-    Model, ModelView, ModelSQL, MatchMixin, ValueMixin, DeactivableMixin,
-    fields, sequence_ordered, tree)
-from trytond import backend
-from trytond.cache import Cache
-from trytond.pyson import Eval, If
-from trytond.transaction import Transaction
-from trytond.pool import Pool, PoolMeta
-from trytond.tools import grouped_slice
-from trytond.tools.multivalue import migrate_property
-
+    DeactivableMixin, Index, MatchMixin, Model, ModelSQL, ModelView, fields,
+    sequence_ordered, tree)
 from trytond.modules.product import price_digits, round_price
+from trytond.pool import Pool
+from trytond.pyson import Eval, If, TimeDelta
+from trytond.tools import grouped_slice
+from trytond.transaction import Transaction, inactive_records
 
 from .exceptions import LocationValidationError
 
@@ -43,14 +40,13 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
 
     name = fields.Char("Name", size=None, required=True, translate=True)
     code = fields.Char(
-        "Code", size=None, select=True,
+        "Code",
         help="The internal identifier used for the location.")
     address = fields.Many2One(
         'party.address', "Address",
         states={
             'invisible': Eval('type') != 'warehouse',
-            },
-        depends=['type'])
+            })
     type = fields.Selection([
         ('supplier', 'Supplier'),
         ('customer', 'Customer'),
@@ -60,17 +56,14 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         ('production', 'Production'),
         ('drop', 'Drop'),
         ('view', 'View'),
-        ], "Location type")
+        ], "Type")
     type_string = type.translated('type')
-    parent = fields.Many2One("stock.location", "Parent", select=True,
+    parent = fields.Many2One(
+        "stock.location", "Parent", ondelete='CASCADE',
         left="left", right="right",
-        states={
-            'invisible': Eval('type') == 'warehouse',
-            },
-        depends=['type'],
         help="Used to add structure above the location.")
-    left = fields.Integer('Left', required=True, select=True)
-    right = fields.Integer('Right', required=True, select=True)
+    left = fields.Integer('Left', required=True)
+    right = fields.Integer('Right', required=True)
     childs = fields.One2Many("stock.location", "parent", "Children",
         help="Used to add structure below the location.")
     flat_childs = fields.Boolean(
@@ -85,13 +78,15 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             'required': Eval('type') == 'warehouse',
             },
         domain=[
-            ('type', '=', 'storage'),
             ['OR',
-                ('parent', 'child_of', [Eval('id')]),
+                ('type', '=', 'storage'),
+                ('id', '=', Eval('storage_location', -1)),
+                ],
+            ['OR',
+                ('parent', 'child_of', [Eval('id', -1)]),
                 ('parent', '=', None),
                 ],
             ],
-        depends=['type', 'id'],
         help="Where incoming stock is received.")
     output_location = fields.Many2One(
         "stock.location", "Output", states={
@@ -99,11 +94,13 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             'required': Eval('type') == 'warehouse',
         },
         domain=[
-            ('type', '=', 'storage'),
             ['OR',
-                ('parent', 'child_of', [Eval('id')]),
+                ('type', '=', 'storage'),
+                ('id', '=', Eval('storage_location', -1)),
+                ],
+            ['OR',
+                ('parent', 'child_of', [Eval('id', -1)]),
                 ('parent', '=', None)]],
-        depends=['type', 'id'],
         help="Where outgoing stock is sent from.")
     storage_location = fields.Many2One(
         "stock.location", "Storage", states={
@@ -113,9 +110,8 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         domain=[
             ('type', 'in', ['storage', 'view']),
             ['OR',
-                ('parent', 'child_of', [Eval('id')]),
+                ('parent', 'child_of', [Eval('id', -1)]),
                 ('parent', '=', None)]],
-        depends=['type', 'id'],
         help="The top level location where stock is stored.")
     picking_location = fields.Many2One(
         'stock.location', 'Picking', states={
@@ -125,19 +121,16 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             ('type', '=', 'storage'),
             ('parent', 'child_of', [Eval('storage_location', -1)]),
             ],
-        depends=['type', 'storage_location'],
         help="Where stock is picked from.\n"
         "Leave empty to use the storage location.")
     lost_found_location = fields.Many2One(
         'stock.location', "Lost and Found",
         states={
             'invisible': Eval('type') != 'warehouse',
-            'readonly': ~Eval('active'),
             },
         domain=[
             ('type', '=', 'lost_found'),
             ],
-        depends=['type', 'active'],
         help="Used, by inventories, when correcting stock levels "
         "in the warehouse.")
     waste_locations = fields.Many2Many(
@@ -148,7 +141,6 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         domain=[
             ('type', '=', 'lost_found'),
             ],
-        depends=['type'],
         help="The locations used for waste products from the warehouse.")
     waste_warehouses = fields.Many2Many(
         'stock.location.waste', 'location', 'warehouse', "Waste Warehouses",
@@ -158,26 +150,29 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         domain=[
             ('type', '=', 'warehouse'),
             ],
-        depends=['type'],
         help="The warehouses that use the location for waste products.")
 
+    allow_pickup = fields.Boolean(
+        "Allow Pickup",
+        states={
+            'invisible': (
+                (Eval('type') != 'warehouse')
+                & ~Eval('address')),
+            })
+
     quantity = fields.Function(
         fields.Float(
-            "Quantity", digits=(16, Eval('quantity_uom_digits', 2)),
-            depends=['quantity_uom_digits'],
+            "Quantity", digits='quantity_uom',
             help="The amount of stock in the location."),
         'get_quantity', searcher='search_quantity')
     forecast_quantity = fields.Function(
         fields.Float(
-            "Forecast Quantity", digits=(16, Eval('quantity_uom_digits', 2)),
-            depends=['quantity_uom_digits'],
+            "Forecast Quantity", digits='quantity_uom',
             help="The amount of stock expected to be in the location."),
         'get_quantity', searcher='search_quantity')
     quantity_uom = fields.Function(fields.Many2One(
-            'product.uom', "Quantity UOM"),
-        'get_quantity_uom')
-    quantity_uom_digits = fields.Function(fields.Integer(
-            "Quantity UOM Digits"),
+            'product.uom', "Quantity UoM",
+            help="The Unit of Measure for the quantities."),
         'get_quantity_uom')
     cost_value = fields.Function(fields.Numeric(
             "Cost Value", digits=price_digits,
@@ -186,7 +181,16 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
 
     @classmethod
     def __setup__(cls):
+        cls.code.search_unaccented = False
         super(Location, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.code, Index.Similarity())),
+                Index(
+                    t,
+                    (t.left, Index.Range()),
+                    (t.right, Index.Range())),
+                })
         cls._order.insert(0, ('name', 'ASC'))
 
         parent_domain = [
@@ -208,7 +212,6 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                     ('type', 'in', childs_mapping[type_]), ()))
         cls.parent.domain = parent_domain
         cls.childs.domain = childs_domain
-        cls.childs.depends.extend(['flat_childs', 'type'])
 
     @classmethod
     def _parent_domain(cls):
@@ -221,7 +224,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             'lost_found': ['lost_found'],
             'view': ['warehouse', 'view', 'storage'],
             'storage': ['warehouse', 'view', 'storage'],
-            'warehouse': [''],
+            'warehouse': ['view'],
             }
 
     @classmethod
@@ -234,26 +237,22 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         return childs_domain
 
     @classmethod
-    def __register__(cls, module_name):
-        super(Location, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-        table.index_action(['left', 'right'], 'add')
-
-    @classmethod
-    def validate(cls, locations):
-        super(Location, cls).validate(locations)
+    def validate_fields(cls, locations, field_names):
+        super().validate_fields(locations, field_names)
         inactives = []
         for location in locations:
-            location.check_type_for_moves()
-            if not location.active:
+            location.check_type_for_moves(field_names)
+            if 'active' in field_names and not location.active:
                 inactives.append(location)
         cls.check_inactive(inactives)
 
-    def check_type_for_moves(self):
+    def check_type_for_moves(self, field_names=None):
         """ Check locations with moves have types compatible with moves. """
+        pool = Pool()
+        Move = pool.get('stock.move')
+        if field_names and 'type' not in field_names:
+            return
         invalid_move_types = ['warehouse', 'view']
-        Move = Pool().get('stock.move')
         if self.type in invalid_move_types:
             # Use root to compute for all companies
             with Transaction().set_user(0):
@@ -263,7 +262,8 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                             ('from_location', '=', self.id),
                             ],
                         ('state', 'not in', ['staging', 'draft']),
-                        ])
+                        ],
+                    order=[], limit=1)
             if moves:
                 raise LocationValidationError(
                     gettext('stock.msg_location_invalid_type_for_moves',
@@ -285,6 +285,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
     def get_empty_locations(cls, locations=None):
         pool = Pool()
         Move = pool.get('stock.move')
+        Product = pool.get('product.product')
         if locations is None:
             locations = cls.search([])
         if not locations:
@@ -292,15 +293,17 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         location_ids = list(map(int, locations))
         # Use root to compute for all companies
         # and ensures inactive locations are in the query
-        with Transaction().set_user(0), \
-                Transaction().set_context(active_test=False):
+        with Transaction().set_user(0), inactive_records():
             query = Move.compute_quantities_query(
                 location_ids, with_childs=True)
             quantities = Move.compute_quantities(
                 query, location_ids, with_childs=True)
             empty = set(location_ids)
+            product_ids = [q[1] for q in quantities.keys()]
+            consumables = {
+                p.id for p in Product.browse(product_ids) if p.consumable}
             for (location_id, product), quantity in quantities.items():
-                if quantity:
+                if quantity and product not in consumables:
                     empty.discard(location_id)
             for sub_ids in grouped_slice(list(empty)):
                 sub_ids = list(sub_ids)
@@ -334,11 +337,11 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
 
     @classmethod
     def check_xml_record(cls, records, values):
-        return True
+        pass
 
     def get_warehouse(self, name):
         # Order by descending left to get the first one in the tree
-        with Transaction().set_context(active_test=False):
+        with inactive_records():
             locations = self.search([
                     ('parent', 'parent_of', [self.id]),
                     ('type', '=', 'warehouse'),
@@ -467,10 +470,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             template = Template(context['product_template'])
             uom = template.default_uom
         if uom:
-            if name == 'quantity_uom':
-                value = uom.id
-            elif name == 'quantity_uom_digits':
-                value = uom.digits
+            value = uom.id
         return {l.id: value for l in locations}
 
     @classmethod
@@ -574,8 +574,16 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                             warehouse=warehouse.rec_name))
 
     @classmethod
-    def delete(cls, *args):
-        super().delete(*args)
+    def delete(cls, locations):
+        # Delete also required children as CASCADING is done separately
+        extra_locations = []
+        for location in locations:
+            extra_locations.extend(filter(None, [
+                        location.input_location,
+                        location.output_location,
+                        location.storage_location,
+                        ]))
+        super().delete(locations + extra_locations)
         cls._default_warehouse_cache.clear()
 
     @classmethod
@@ -644,86 +652,11 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             ]
 
 
-supplier_location = fields.Many2One(
-    'stock.location', "Supplier Location", domain=[('type', '=', 'supplier')],
-    help="The default source location for stock received from the party.")
-customer_location = fields.Many2One(
-    'stock.location', "Customer Location", domain=[('type', '=', 'customer')],
-    help="The default destination location for stock sent to the party.")
-
-
-class Party(metaclass=PoolMeta):
-    __name__ = 'party.party'
-    supplier_location = fields.MultiValue(supplier_location)
-    customer_location = fields.MultiValue(customer_location)
-    locations = fields.One2Many(
-        'party.party.location', 'party', "Locations")
-
-    @classmethod
-    def multivalue_model(cls, field):
-        pool = Pool()
-        if field in {'supplier_location', 'customer_location'}:
-            return pool.get('party.party.location')
-        return super(Party, cls).multivalue_model(field)
-
-    @classmethod
-    def default_supplier_location(cls, **pattern):
-        return cls.multivalue_model(
-            'supplier_location').default_supplier_location()
-
-    @classmethod
-    def default_customer_location(cls, **pattern):
-        return cls.multivalue_model(
-            'customer_location').default_customer_location()
-
-
-class PartyLocation(ModelSQL, ValueMixin):
-    "Party Location"
-    __name__ = 'party.party.location'
-    party = fields.Many2One(
-        'party.party', "Party", ondelete='CASCADE', select=True)
-    supplier_location = supplier_location
-    customer_location = customer_location
-
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(PartyLocation, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.extend(['supplier_location', 'customer_location'])
-        value_names.extend(['supplier_location', 'customer_location'])
-        migrate_property(
-            'party.party', field_names, cls, value_names,
-            parent='party', fields=fields)
-
-    @classmethod
-    def default_supplier_location(cls):
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        try:
-            return ModelData.get_id('stock', 'location_supplier')
-        except KeyError:
-            return None
-
-    @classmethod
-    def default_customer_location(cls):
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        try:
-            return ModelData.get_id('stock', 'location_customer')
-        except KeyError:
-            return None
-
-
 class ProductsByLocationsContext(ModelView):
     'Products by Locations'
     __name__ = 'stock.products_by_locations.context'
+
+    company = fields.Many2One('company.company', "Company", required=True)
     forecast_date = fields.Date(
         'At Date',
         help="The date for which the stock quantity is calculated.\n"
@@ -732,6 +665,10 @@ class ProductsByLocationsContext(ModelView):
     stock_date_end = fields.Function(fields.Date('At Date'),
         'on_change_with_stock_date_end')
 
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
     @staticmethod
     def default_forecast_date():
         Date_ = Pool().get('ir.date')
@@ -750,20 +687,16 @@ class ProductsByLocations(DeactivableMixin, ModelSQL, ModelView):
 
     product = fields.Many2One('product.product', "Product")
     quantity = fields.Function(
-        fields.Float(
-            "Quantity", digits=(16, Eval('default_uom_digits', 2)),
-            depends=['default_uom_digits']),
+        fields.Float("Quantity", digits='default_uom'),
         'get_product', searcher='search_product')
     forecast_quantity = fields.Function(
-        fields.Float(
-            "Forecast Quantity", digits=(16, Eval('default_uom_digits', 2)),
-            depends=['default_uom_digits']),
+        fields.Float("Forecast Quantity", digits='default_uom'),
         'get_product', searcher='search_product')
     default_uom = fields.Function(
-        fields.Many2One('product.uom', "Default UOM"),
+        fields.Many2One(
+            'product.uom', "Default UoM",
+            help="The default Unit of Measure."),
         'get_product', searcher='search_product')
-    default_uom_digits = fields.Function(
-        fields.Integer("Default UOM Digits"), 'get_product')
     cost_value = fields.Function(
         fields.Numeric("Cost Value"), 'get_product')
     consumable = fields.Function(
@@ -791,6 +724,13 @@ class ProductsByLocations(DeactivableMixin, ModelSQL, ModelView):
                 columns.append(column.as_(fname))
         return product.select(*columns)
 
+    def get_rec_name(self, name):
+        return self.product.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('product.rec_name',) + tuple(clause[1:])]
+
     def get_product(self, name):
         value = getattr(self.product, name)
         if isinstance(value, Model):
@@ -799,8 +739,8 @@ class ProductsByLocations(DeactivableMixin, ModelSQL, ModelView):
 
     @classmethod
     def search_product(cls, name, clause):
-        nested = clause[0].lstrip(name)
-        return [('product.' + name + nested,) + tuple(clause[1:])]
+        nested = clause[0][len(name):]
+        return [('product.' + name + nested, *clause[1:])]
 
 
 class LocationLeadTime(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
@@ -811,13 +751,20 @@ class LocationLeadTime(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         ondelete='CASCADE',
         domain=[
             ('type', '=', 'warehouse'),
+            ('id', '!=', Eval('warehouse_to', -1)),
             ])
     warehouse_to = fields.Many2One('stock.location', 'Warehouse To',
         ondelete='CASCADE',
         domain=[
             ('type', '=', 'warehouse'),
+            ('id', '!=', Eval('warehouse_from', -1)),
             ])
-    lead_time = fields.TimeDelta('Lead Time',
+    lead_time = fields.TimeDelta(
+        "Lead Time",
+        domain=['OR',
+            ('lead_time', '=', None),
+            ('lead_time', '>=', TimeDelta()),
+            ],
         help="The time it takes to move stock between the warehouses.")
 
     @classmethod
