@@ -1,15 +1,15 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from itertools import chain
+from itertools import chain, groupby
 
-from sql import Literal, For
+from sql import For, Literal
 
 from trytond.i18n import gettext
-from trytond.model import Workflow, ModelView, ModelSQL, fields
-from trytond.pyson import Eval
-from trytond.transaction import Transaction
+from trytond.model import Index, ModelSQL, ModelView, Workflow, fields
 from trytond.pool import Pool
+from trytond.pyson import Eval
 from trytond.tools import grouped_slice
+from trytond.transaction import Transaction
 
 from .exceptions import PeriodCloseError
 
@@ -19,7 +19,7 @@ class Period(Workflow, ModelSQL, ModelView):
     __name__ = 'stock.period'
     date = fields.Date('Date', required=True, states={
             'readonly': Eval('state') == 'closed',
-            }, depends=['state'],
+            },
         help="When the stock period ends.")
     company = fields.Many2One(
         'company.company', "Company", required=True,
@@ -29,12 +29,19 @@ class Period(Workflow, ModelSQL, ModelView):
     state = fields.Selection([
         ('draft', 'Draft'),
         ('closed', 'Closed'),
-        ], 'State', select=True, readonly=True,
+        ], "State", readonly=True, sort=False,
         help="The current state of the stock period.")
 
     @classmethod
     def __setup__(cls):
         super(Period, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.company, Index.Equality()),
+                (t.date, Index.Range(order='DESC')),
+                where=t.state == 'closed'))
         cls._transitions |= set((
                 ('draft', 'closed'),
                 ('closed', 'draft'),
@@ -102,7 +109,7 @@ class Period(Workflow, ModelSQL, ModelView):
         # from a different transaction. It will not be taken in the pbl
         # computation but it is quite rare because only past periods are
         # closed.
-        database.lock(connection, Move._table)
+        Move.lock()
         if database.has_select_for():
             move = Move.__table__()
             query = move.select(Literal(1), for_=For('UPDATE', nowait=True))
@@ -112,12 +119,14 @@ class Period(Workflow, ModelSQL, ModelView):
         locations = Location.search([
                 ('type', 'not in', ['warehouse', 'view']),
                 ], order=[])
-        today = Date.today()
 
-        recent_date = max(period.date for period in periods)
-        if recent_date >= today:
-            raise PeriodCloseError(
-                gettext('stock.msg_period_close_date'))
+        for company, c_periods in groupby(periods, key=lambda p: p.company):
+            with Transaction().set_context(company=company.id):
+                today = Date.today()
+            recent_date = max(period.date for period in c_periods)
+            if recent_date >= today:
+                raise PeriodCloseError(
+                    gettext('stock.msg_period_close_date'))
         if Move.search([
                     ('state', '=', 'assigned'),
                     ['OR', [
@@ -162,10 +171,28 @@ class Cache(ModelSQL, ModelView):
     It is used to store cached computation of stock quantities.
     '''
     __name__ = 'stock.period.cache'
-    period = fields.Many2One('stock.period', 'Period', required=True,
-        readonly=True, select=True, ondelete='CASCADE')
-    location = fields.Many2One('stock.location', 'Location', required=True,
-        readonly=True, select=True, ondelete='CASCADE')
-    product = fields.Many2One('product.product', 'Product', required=True,
-        readonly=True, select=True, ondelete='CASCADE')
+    period = fields.Many2One(
+        'stock.period', "Period",
+        required=True, readonly=True, ondelete='CASCADE')
+    location = fields.Many2One(
+        'stock.location', "Location",
+        required=True, readonly=True, ondelete='CASCADE')
+    product = fields.Many2One(
+        'product.product', "Product",
+        required=True, readonly=True, ondelete='CASCADE')
     internal_quantity = fields.Float('Internal Quantity', readonly=True)
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(
+                    t,
+                    (t.period, Index.Equality()),
+                    (t.product, Index.Equality()),
+                    include=[t.internal_quantity]),
+                Index(
+                    t,
+                    (t.location, Index.Equality())),
+                })
